# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types for PyTree utilities."""

from typing import Any
import jax
import numpy as np
from orbax.checkpoint._src.tree import types

JsonType = types.JsonType

PyTree = Any
PyTreeOf = types.PyTreeOf

PyTreeKey = types.PyTreeKey
PyTreeKeyPath = types.PyTreePath

ScalarType = int | float | bool
LeafType = jax.Array | np.ndarray | str | ScalarType | Any
AbstractLeafType = Any  # TODO(cpgaffney): Add a type for abstract leaves.

JsonType = types.JsonType

PLACEHOLDER = ...
